/**
 * \file activebg.h
 * \brief ActiveBg osztály deklarációja
 *
 */

#ifndef ACTIVEBG_H
#define ACTIVEBG_H

#include "widget.h"

/// Egy olyan grafikusan nem megjelenő widget, amire a callback
/// függvény tehető. A teljes képernyőn aktív.
/// A gomb "megnyomására" meghívódik a paraméterként kapott
/// osztály függvényhívás operátora a beállított int paraméterrel.
/// Azért template, hogy könnyebben alkalmazkodjon a különféle
/// callback függvényekhez, funktorokhoz.
template<class T = void (&)(const SDL_Event&, int) >
class ActiveBg : public Widget {
    T& cb;          ///< callback objektum referencia
    int val;        ///< int érték, amit átad a callback függvénynek
public:
    /// Konstruktor elteszi a callback rutin címét és a híváskor
    /// átadandó int paramétert
    ActiveBg(T& cb, int val = 0) : cb(cb), val(val) { }

    /// Mivel igazi ablak nincs, a teljes képernyő a tárolóhoz tartozik.
    /// Minden pontot magáénak tekint.
    bool contains(const Point& p) {
        return true;
    }
    /// A felhasználói felületen megnyomták a gombot
    bool eventHandler(const SDL_Event& ev) {
        if (ev.type == SDL_MOUSEBUTTONDOWN) {
            if (ev.button.button == SDL_BUTTON_RIGHT)
                cb(ev, val);
            return true;
        }
        return false;
    }
    void draw() const {}
};

#endif // ACTIVEBG_H
