/**
 * \file appshell.cpp
 * \brief AppShell osztály tagfüggvényeinek megvalósítása
 *
 * \bug Érintőképernyőn SDL 1.2-vel az érintésből származó pozíció rossz. 1.3-as változatban ez már javítva van.
 */

#include "appshell.h"
#include "memtrace.h"

/// Eseményhurok.
/// Adminisztrálja az egér mozgását.
/// Minden, a tárolóban levő objektumra megnézi, hogy az esemény érinti-e az objektumot.
/// Ha igen, akkor meghívja az objektum eseménykezelőjét.
/// A tárolással fordított sorrendben keres, így a később betett objektum van "felül".
/// Az objektum eseménykezelője ettől még dönthet úgy, hogy nem kezeli.
/// Kihasználható, hogy az eseménykezelő meghívása előtt mindig van contains hívás.
void AppShell::mainLoop() {
    SDL_RenderPresent(renderer);
    SDL_Event ev;
    Point cursor;
    while (SDL_WaitEvent(&ev) && ev.type != SDL_QUIT) {
        bool changed = false;
        if (ev.type == SDL_MOUSEMOTION) {  // egérmozgás követése
            cursor.x = ev.motion.x;
            cursor.y = ev.motion.y;
        }
        size_t i = widgets.size()-1;
        while (i >= 0 && i < widgets.size()) {      // a méret megváltozhat a ciklusban !
            if (widgets[i]->contains(cursor)) {     // ha a widget területére esik
                if (widgets[i]->eventHandler(ev)) { // akkor elküldi az eseményt
                    changed = true;                 // megjegyzi, hogy változhatott a rajz
                    break;
                }
            }
            i--;
        }
        if (changed) {      // ha változhatott a rajz, frissíti a képernyőt
            draw();
            SDL_RenderPresent(renderer);
        }
    }
}
