/**
 * \file button.h
 * \brief Button osztály deklarációja
 *
 */

#ifndef BUTTON_H
#define BUTTON_H

#include "circle.h"

/// Egy olyan kör, amire a callback függvény tehető.
/// A gomb "megnyomására" meghívódik a paraméterként kapott
/// osztály függvényhívás operátora a beállított int paraméterrel.
/// Azért template, hogy könnyebben alkalmazkodjon a különféle
/// callback függvényekhez, funktorokhoz.
template<class T = void (&)(const SDL_Event&, int) >
class Button : public Circle {
    T& cb;          ///< callback objektum referencia
    int val;        ///< int érték
public:
    /// Konstruktor elteszi a callback rutin címét és a híváskor
    /// átadandó int paramétert
    Button(const Point& p0, int r, T& cb, int val = 0,
            const Color& fg = WHITE, const Color& bg = BLACK) :
        Circle(p0, r, fg, bg, true), cb(cb), val(val) { }

    /// A felhasználói felületen megnyomták a gombot
    bool eventHandler(const SDL_Event& ev) {
        if (ev.type == SDL_MOUSEBUTTONDOWN) {
            if (ev.button.button == SDL_BUTTON_LEFT)
                cb(ev, val);
            return true;
        }
        return Circle::eventHandler(ev);
    }
};

#endif // BUTTON_H
