/**
 *  \file container.h
 *  \brief Container osztály deklarációja
 */

#ifndef CONTAINER_H
#define CONTAINER_H

#include "widget.h"
#include <vector>

/// Általános widget konténer (heterogén kollekció).
class Container : public Widget {
protected:
    std::vector<Widget*> widgets;   ///< Itt tároljuk a pointereket
    Point cursor;                   ///< egérmozgás követésére
public:
    /// add
    /// widget hozzáadása a konténerhez
    /// @param w - widget pointere
    void add(Widget *w);

    /// clear
    /// Minden tárolt tartalmazott objektumot eldob és megszüntet
    void clear();

    /// contains
    /// Elteszi a pozíciót, mert a hívást követi az esemény kezelése.
    /// @param p - adott pont
    /// @return false
    bool contains(const Point& p) {
        cursor = p;     // egérmozgás követése
        return false;   // nem tudjuk a használatot, ezért inkább false
    }

    /// Widget eseménykezelője.
    /// Minden tartalamzott objektumot végigkérdez.
    /// @param ev - SDL eseménystruktúra
    /// @return true, ha lekezelte az eseményt. false, ha továbbadja a hívónak a kezelést
    bool eventHandler(const SDL_Event&);

    /// Minden tartalmazott osztály rajzoló funkcióját meghívja.
    void draw() const;

    /// destruktor
    ~Container() { clear(); }
};

#endif // CONTAINER_H
