/**
 * \file widget.h
 * \brief Widget alaposztály deklarációja
 */

#ifndef WIDGET_H
#define WIDGET_H

#include <SDL2/sdl.h>
#include <SDL2/SDL2_gfxPrimitives.h>
#include "point.h"
#include "color.h"

/// Absztrakt osztály.
/// Statikus tagként tárolja az SDL surface pointert.
/// Nem szép megoldás, de szebb mintha globális lenne.
class Widget {
protected:
     static SDL_Renderer *renderer;  ///< eldugott globális változó
     Point p0;                  ///< Widget origója
     Color fg;                  ///< Előtér szín
     Color bg;                  ///< Háttérszín
public:
    /// Konstruktor
    /// @param p0 - referencia pont
    Widget(const Point& p0 = Point(),
           const Color& fg = WHITE,
           const Color& bg = BLACK) :p0(p0), fg(fg), bg(bg) {}

    /// renderer beállítása
    /// @param r - SDL remderer
    static void setRenderer(SDL_Renderer* r) { renderer = r; }

    /// contains
    /// Egy adott pontról eldönti, hogy a widget területére esik-e
    /// @param p - adott pont
    /// @return true, ha a pont az widgeten belül van. false egyébként
    virtual bool contains(const Point& p) { return false; }

    /// Widget eseménykezelője
    /// @param ev - SDL eseménystruktúra
    /// @return true, ha lekezelte az eseményt. false, ha továbbadja a hívónak a kezelést
    virtual bool eventHandler(const SDL_Event& ev) { return false; };

    /// move: origóját eltolja az új pontba
    /// @param d - eltolás vektora
    void move(const Point& d);

    /// draw
    /// Felrajzolja a widgetet
    /// Tipikusan az eseménykezelő hívja, de esetenként más is.
    virtual void draw() const = 0;      // tisztán virtuális

    /// destruktor
    virtual ~Widget() {}
};

#endif // WIDGET_H
