/**
 * \file ctor_dtor2.cpp
 *
 * Egyszerű osztályhierarchia az ctor/dtor hívási sorrendjének bemutatására.
 * Az azonos feladatok végrehajtási sorrendje (pl. adattagok konstruktorhívása,
 * vagy ősosztályok konstruktorának hívási sorrendje) a deklaráció sorrendjétől
 * függ.
 */

#include <iostream>

/// Korábban használt makró, ami kiírja a paramétereit, majd változtatás nélkül kigenerálja azokat.
/// A kiírást egy tabulátorral kezdi és soremeléssel zárja.
/// Használatával elérhető, hogy a paraméterként adott utasítások, deklarációk a végrehajtás előtt
/// kiíródjanak a képernyőre. Így hatásuk jól megfigyelhető.
#define _(...)  std::cout << "\t" << #__VA_ARGS__ << std::endl; __VA_ARGS__

using std::cout;
using std::cerr;
using std::cin;
using std::endl;

/**
 Konstruktor:
 * Öröklési lánc végén hívja a virtuális alaposztályok konstruktorait.
 * Hívja a közvetlen, nem virtuális alaposztályok konstruktorait.
 * Létrehozza a saját részt:
   - beállítja a virtuális alaposztály mutatóit
   - beállítja a virtuális függvények mutatóit
   - hívja a tartalmazott objektumok konstruktorait
   - végrehajtja a programozott törzset
 */


/**
 Destruktor:
 * Megszünteti a saját részt:
   - végrehajtja a programozott törzset
   - hívja a tartalmazott objektumok destruktorait
   - beállítja a virtuális függvények mutatóit
   - beállítja a virtuális alaposztály mutatóit
 * Hívja a közvetlen, nem virtuális alaposztályok destruktorait.
 * Öröklési lánc végén hívja a virtuális alaposztályok destruktorait.
 */


/// Virtuális alaposztályok
struct VirtAlap1 {
    VirtAlap1() {
        cout << "ctor: VirtAlap1" << endl;
    }
    ~VirtAlap1() {
        cout << "dtor: VirtAlap1" << endl;
    }
};

struct VirtAlap2 {
    VirtAlap2() {
        cout << "ctor: VirtAlap2" << endl;
    }
    ~VirtAlap2() {
        cout << "dtor: VirtAlap2" << endl;
    }
};


/// Nem virt alaposztályok
struct NemVirtAlap1 : virtual  VirtAlap1, virtual VirtAlap2 {
    NemVirtAlap1()  {
        cout << "ctor: NemVirtAlap1" << endl;
    }
    ~NemVirtAlap1()  {
        cout << "dtor: NemVirtAlap1" << endl;
    }
};

struct NemVirtAlap2 {
    NemVirtAlap2()  {
        cout << "ctor: NemVirtAlap2" << endl;
    }
    ~NemVirtAlap2()  {
        cout << "dtor: NemVirtAlap2" << endl;
    }
};

/// Adattagok
struct Adat1 {
    Adat1()  { cout << "ctor: Adat1" << endl; }
    ~Adat1() { cout << "dtor: Adat1" << endl; }
};

struct Adat2 {
    Adat2()  { cout << "ctor: Adat2" << endl; }
    ~Adat2() { cout << "dtor: Adat2" << endl; }
};

/// Az OSZTALY
struct A : NemVirtAlap1, NemVirtAlap2 {
    Adat1 adat1;
    Adat2 adat2;
    A()  {
        cout << "ctor: A" << endl;
    }
    ~A()  {
        cout << "dtor: A" << endl;
    }
};

int main() {
_(  A* p = new A;   )
    cout << "Tovabb: ENTER" << endl;
    cin.ignore(100, '\n');
_(  delete p;       )
    return 0;
}

